@@ -60,6 +60,7 @@ from dkim.canonicalization import Relaxed as RelaxedCanonicalization
 from dkim.crypto import (
     DigestTooLargeError,
     HASH_ALGORITHMS,
+    ARC_HASH_ALGORITHMS,
     parse_pem_private_key,
     parse_public_key,
     RSASSA_PKCS1_v1_5_sign,
@@ -109,6 +110,7 @@ CV_Pass = b'pass'
 CV_Fail = b'fail'
 CV_None = b'none'
 
+
 class HashThrough(object):
   def __init__(self, hasher):
     self.data = []
@@ -128,45 +130,56 @@ class HashThrough(object):
   def hashed(self):
     return b''.join(self.data)
 
+
 def bitsize(x):
     """Return size of long in bits."""
     return len(bin(x)) - 2
 
+
 class DKIMException(Exception):
     """Base class for DKIM errors."""
     pass
 
+
 class InternalError(DKIMException):
     """Internal error in dkim module. Should never happen."""
     pass
 
+
 class KeyFormatError(DKIMException):
     """Key format error while parsing an RSA public or private key."""
     pass
 
+
 class MessageFormatError(DKIMException):
     """RFC822 message format error."""
     pass
 
+
 class ParameterError(DKIMException):
     """Input parameter error."""
     pass
 
+
 class ValidationError(DKIMException):
     """Validation error."""
     pass
 
+
 class AuthresNotFoundError(DKIMException):
     """ Authres Package not installed, needed for ARC """
     pass
 
+
 class NaClNotFoundError(DKIMException):
     """ Nacl package not installed, needed for ed25119 signatures """
     pass
 
+
 class UnknownKeyTypeError(DKIMException):
     """ Key type (k tag) is not known (rsa/ed25519) """
 
+
 def select_headers(headers, include_headers):
     """Select message header fields to be signed/verified.
 
@@ -192,10 +205,12 @@ def select_headers(headers, include_headers):
         lastindex[h] = i
     return sign_headers
 
+
 # FWS  =  ([*WSP CRLF] 1*WSP) /  obs-FWS ; Folding white space  [RFC5322]
 FWS = br'(?:(?:\s*\r?\n)?\s+)?'
 RE_BTAG = re.compile(br'([;\s]b'+FWS+br'=)(?:'+FWS+br'[a-zA-Z0-9+/=])*(?:\r?\n\Z)?')
 
+
 def hash_headers(hasher, canonicalize_headers, headers, include_headers,
                  sigheader, sig):
     """Update hash for signed message header fields."""
@@ -212,6 +227,7 @@ def hash_headers(hasher, canonicalize_headers, headers, include_headers,
         hasher.update(y)
     return sign_headers
 
+
 def hash_headers_ed25519(pk, canonicalize_headers, headers, include_headers,
                  sigheader, sig):
     """Update hash for signed message header fields."""
@@ -227,6 +243,7 @@ def hash_headers_ed25519(pk, canonicalize_headers, headers, include_headers,
         hash_header += x + y
     return sign_headers, hash_header
 
+
 def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'd', b'h', b's'], arc=False):
     """Validate DKIM or ARC Signature fields.
     Basic checks for presence and correct formatting of mandatory fields.
@@ -235,11 +252,15 @@ def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'
     @param mandatory_fields: A list of non-optional fields
     @param arc: flag to differentiate between dkim & arc
     """
+    if arc:
+        hashes = ARC_HASH_ALGORITHMS
+    else:
+        hashes = HASH_ALGORITHMS
     for field in mandatory_fields:
         if field not in sig:
             raise ValidationError("missing %s=" % field)
 
-    if b'a' in sig and not sig[b'a'] in HASH_ALGORITHMS:
+    if b'a' in sig and not sig[b'a'] in hashes:
         raise ValidationError("unknown signature algorithm: %s" % sig[b'a'])
 
     if b'b' in sig:
@@ -299,6 +320,7 @@ def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'
                     "x= value is less than t= value (x=%s t=%s)" %
                     (sig[b'x'], sig[b't']))
 
+
 def rfc822_parse(message):
     """Parse a message in RFC822 format.
 
@@ -327,6 +349,7 @@ def rfc822_parse(message):
         i += 1
     return (headers, b"\r\n".join(lines[i:]))
 
+
 def text(s):
     """Normalize bytes/str to str for python 2/3 compatible doctests.
     >>> text(b'foo')
@@ -341,6 +364,7 @@ def text(s):
     if type(s) is str: return s
     return s.encode('ascii')
 
+
 def fold(header, namelen=0):
     """Fold a header line into multiple crlf-separated lines at column 72.
 
@@ -388,6 +412,7 @@ def fold(header, namelen=0):
         else:
             return pre + header
 
+
 def load_pk_from_dns(name, dnsfunc=get_txt):
   s = dnsfunc(name)
   if not s:
@@ -424,6 +449,7 @@ def load_pk_from_dns(name, dnsfunc=get_txt):
       raise KeyFormatError('unknown algorithm in k= tag: {0}'.format(pub[b'k']))
   return pk, keysize, ktag
 
+
 #: Abstract base class for holding messages and options during DKIM/ARC signing and verification.
 class DomainSigner(object):
   # NOTE - the first 2 indentation levels are 2 instead of 4
@@ -460,12 +486,12 @@ class DomainSigner(object):
   #:
   #: The short list below is the result more of instinct than logic.
   #: @since: 0.5
-  FROZEN = (b'from',b'date',b'subject')
+  FROZEN = (b'from',)
 
   #: The rfc6376 recommended header fields to sign
   #: @since: 0.5
   SHOULD = (
-    b'sender', b'reply-to', b'subject', b'date', b'message-id', b'to', b'cc',
+    b'from', b'sender', b'reply-to', b'subject', b'date', b'message-id', b'to', b'cc',
     b'mime-version', b'content-type', b'content-transfer-encoding',
     b'content-id', b'content-description', b'resent-date', b'resent-from',
     b'resent-sender', b'resent-to', b'resent-cc', b'resent-message-id',
@@ -500,11 +526,30 @@ class DomainSigner(object):
     >>> dkim = DKIM()
     >>> dkim.add_frozen(DKIM.RFC5322_SINGLETON)
     >>> [text(x) for x in sorted(dkim.frozen_sign)]
-    ['cc', 'date', 'from', 'in-reply-to', 'message-id', 'references', 'reply-to', 'sender', 'subject', 'to']
+    ['cc', 'date', 'from', 'in-reply-to', 'message-id', 'references', 'reply-to', 'sender', 'to']
+    >>> dkim2 = DKIM()
+    >>> dkim2.add_frozen((b'date',b'subject'))
+    >>> [text(x) for x in sorted(dkim2.frozen_sign)]
+    ['date', 'from', 'subject']
     """
     self.frozen_sign.update(x.lower() for x in s
         if x.lower() not in self.should_not_sign)
 
+
+  def add_should_not(self,s):
+    """ Add headers not in should_not_sign to frozen_sign.
+    @param s: list of headers to add to frozen_sign
+    @since: 0.9
+
+    >>> dkim = DKIM()
+    >>> dkim.add_should_not(DKIM.RFC5322_SINGLETON)
+    >>> [text(x) for x in sorted(dkim.should_not_sign)]
+    ['bcc', 'cc', 'comments', 'date', 'dkim-signature', 'in-reply-to', 'keywords', 'message-id', 'received', 'references', 'reply-to', 'resent-bcc', 'return-path', 'sender', 'to']
+    """
+    self.should_not_sign.update(x.lower() for x in s
+        if x.lower() not in self.frozen_sign)
+
+
   #: Load a new message to be signed or verified.
   #: @param message: an RFC822 formatted message to be signed or verified
   #: (with either \\n or \\r\\n line endings)
@@ -668,6 +713,7 @@ class DomainSigner(object):
     else:
         raise UnknownKeyTypeError(ktag)
 
+
 #: Hold messages and options during DKIM signing and verification.
 class DKIM(DomainSigner):
   #: Sign an RFC822 message and return the DKIM-Signature header line.
@@ -807,6 +853,7 @@ class DKIM(DomainSigner):
 
     return self.verify_sig(sig, include_headers, sigheaders[idx], dnsfunc)
 
+
 #: Hold messages and options during ARC signing and verification.
 class ARC(DomainSigner):
   #: Header fields used by ARC
@@ -871,6 +918,8 @@ class ARC(DomainSigner):
   def sign(self, selector, domain, privkey, srv_id, include_headers=None,
            timestamp=None, standardize=False):
 
+    INSTANCE_LIMIT = 50 # Maximum allowed i= value
+    self.add_should_not(('Authentication-Results',))
     # check if authres has been imported
     try:
         AuthenticationResultsHeader
@@ -933,6 +982,8 @@ class ARC(DomainSigner):
     instance = 1
     if len(arc_headers_w_instance) != 0:
         instance = max_instance + 1
+    if instance > INSTANCE_LIMIT:
+        raise ParameterError("Maximum instance tag value exceeded")
 
     if instance == 1 and chain_validation_status != CV_None:
         raise ParameterError("No existing chain found on message, cv should be none")
@@ -1156,10 +1207,12 @@ class ARC(DomainSigner):
     self.logger.debug("as valid: %r" % as_valid)
     return output
 
+
 def sign(message, selector, domain, privkey, identity=None,
          canonicalize=(b'relaxed', b'simple'),
          signature_algorithm=b'rsa-sha256',
          include_headers=None, length=False, logger=None):
+    # type: (bytes, bytes, bytes, bytes, bytes, tuple, bytes, list, bool, any) -> bytes
     """Sign an RFC822 message and return the DKIM-Signature header line.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param selector: the DKIM selector value for the signature
@@ -1178,12 +1231,14 @@ def sign(message, selector, domain, privkey, identity=None,
     d = DKIM(message,logger=logger,signature_algorithm=signature_algorithm)
     return d.sign(selector, domain, privkey, identity=identity, canonicalize=canonicalize, include_headers=include_headers, length=length)
 
+
 def verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
     """Verify the first (topmost) DKIM signature on an RFC822 formatted message.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param logger: a logger to which debug info will be written (default None)
     @return: True if signature verifies or False otherwise
     """
+    # type: (bytes, any, function, int) -> bool
     d = DKIM(message,logger=logger,minkey=minkey)
     try:
         return d.verify(dnsfunc=dnsfunc)
@@ -1196,10 +1251,12 @@ def verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
 dkim_sign = sign
 dkim_verify = verify
 
+
 def arc_sign(message, selector, domain, privkey,
              srv_id, signature_algorithm=b'rsa-sha256',
              include_headers=None, timestamp=None,
              logger=None, standardize=False):
+    # type: (bytes, bytes, bytes, bytes, bytes, bytes, list, any, any, bool) -> list
     """Sign an RFC822 message and return the ARC set header lines for the next instance
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param selector: the DKIM selector value for the signature
@@ -1208,18 +1265,21 @@ def arc_sign(message, selector, domain, privkey,
     @param srv_id: the authserv_id used to identify the ADMD's AR headers
     @param signature_algorithm: the signing algorithm to use when signing
     @param include_headers: a list of strings indicating which headers are to be signed (default all headers not listed as SHOULD NOT sign)
+    @param timestamp: the time in integer seconds when the message is sealed (default is int(time.time) based on platform, can be string or int)
     @param logger: a logger to which debug info will be written (default None)
     @return: A list containing the ARC set of header fields for the next instance
     @raise DKIMException: when the message, include_headers, or key are badly formed.
     """
 
-    a = ARC(message,logger=logger,signature_algorithm=signature_algorithm)
+    a = ARC(message,logger=logger,signature_algorithm=b'rsa-sha256')
     if not include_headers:
         include_headers = a.default_sign_headers()
     return a.sign(selector, domain, privkey, srv_id, include_headers=include_headers,
                   timestamp=timestamp, standardize=standardize)
 
+
 def arc_verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
+    # type: (bytes, any, function, int) -> tuple
     """Verify the ARC chain on an RFC822 formatted message.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param logger: a logger to which debug info will be written (default None)
